from __future__ import unicode_literals
import datetime

from django.conf import settings
from django.contrib.auth.models import User
import redis

from rapidsms.tests.harness import RapidTest

from .. import rolodex


class XrayTestBase(RapidTest):

    def setUp(self):
        self.prefix = "testxray"
        self.rolodex = rolodex.Rolodex(prefix=self.prefix)
        self.bitmapist_db = redis.Redis(host='localhost', port=6379, db=3)
        self.rolodex_db = redis.Redis(host='localhost', port=6379, db=4)
        self.cleaver_db = redis.Redis(host='localhost', port=6379, db=5)
        # Before doing anything else, we must clear out the dummy backend
        # as this is not automatically flushed between tests.
        self.clear_test_data()
        return super(XrayTestBase, self).setUp()

    def clear_test_data(self):
        # clear cleaver test data
        experiment_keys = self.cleaver_db.keys("%s:*" % self.prefix)
        for experiment_key in experiment_keys:
            self.experiment_db.delete(experiment_key)

        # clear rolodex test data
        rolodex_keys = self.rolodex_db.keys("%s:*" % self.prefix)
        for rolodex_key in rolodex_keys:
            self.rolodex_db.delete(rolodex_key)

        # clear bitmapist test data
        event_keys = self.bitmapist_db.keys("trackist_%s*" % self.prefix)
        for event_key in event_keys:
            self.bitmapist_db.delete(event_key)

        op_keys = self.bitmapist_db.keys("trackist_bitop_AND_trackist_%s*" %
                                         self.prefix)
        for op_key in op_keys:
            self.redis.bitmapist_db(op_key)

    def create_reporter(self, **kwargs):
        defaults = {
            'name': self.random_string(25),
            'birth_date': datetime.date(2010, 3, 14),
            'sex': 'M',
        }
        defaults.update(**kwargs)

        return defaults

    def create_user(self, username=None, password=None, email=None,
                    user_permissions=None, groups=None, **kwargs):
        username = username or self.random_string(25)
        password = password or self.random_string(25)
        email = email or '{0}@example.com'.format(self.random_string(25))
        user = User.objects.create_user(username, email, password)
        if user_permissions:
            user.user_permissions = user_permissions
        if groups:
            user.groups = groups
        if kwargs:
            User.objects.filter(pk=user.pk).update(**kwargs)
        return User.objects.get(pk=user.pk)



from __future__ import unicode_literals
#from collections import defaultdict
import datetime
import json

from django.template import RequestContext
from django.shortcuts import render_to_response
from django.http import HttpResponse
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.decorators import permission_required
from django.contrib.auth.decorators import login_required

from cleaver.experiment import VariantStat
from cleaver.backend.redis import RedisBackend
from bitmapist import cohort

from .events import Tracker
from .events import WebWeekEvents
from .events import SMSWeekEvents

# TODO use project config?
backend = RedisBackend()
tracker = Tracker()


@login_required
@permission_required('xray.view_events', raise_exception=True)
def index(request):
    cleaver = request.environ['cleaver']
    tracker.web_event('view_xray_dashboard', cleaver.identity)
    tracker.web_event('active', cleaver.identity)

    num_experiments = len(backend.all_experiments())
    now = datetime.datetime.utcnow()

    num_web_active_this_week = len(WebWeekEvents('active', now.year,
                                                 now.isocalendar()[1]))
    num_sms_active_this_week = len(SMSWeekEvents('active', now.year,
                                                 now.isocalendar()[1]))
    return render_to_response(
        "xray/index.html",
        {"experiments": num_experiments,
         "web_active": num_web_active_this_week,
         "sms_active": num_sms_active_this_week},
        context_instance=RequestContext(request))


def format_percentage(f):
    return '{:.2%}'.format(f) if f else None


def _experiment_data(experiment):
    control = VariantStat(experiment.control, experiment)
    data = {
        'name': experiment.name,
        'started_on': experiment.started_on,
        'participants': experiment.participants,
        'conversions': experiment.conversions,
        'control_conversion_rate': control.conversion_rate,
    }
    # list of tuples bc template expects control to be first
    variants_data = []
    for variant_name in experiment.variants:
        variant = VariantStat(variant_name, experiment)
        improvement_from_control = None
        if control.conversion_rate > 0:
            improvement_from_control =\
                format_percentage(abs((variant.conversion_rate /
                                       control.conversion_rate) - 1))

        variant_data = {
            'name': variant_name,
            'is_control': (variant_name == experiment.control),
            # experiment.participants_for uses an unbustable cache,
            # so call the backend method directly
            'participants': experiment.backend.participants(experiment.name,
                                                            variant_name),
            # experiment.conversions_for uses an unbustable cache,
            # so call the backend method directly
            'conversions': experiment.backend.conversions(experiment.name,
                                                          variant_name),
            'conversion_rate': format_percentage(variant.conversion_rate),
            'z_score': variant.z_score,
            'confidence_level': variant.confidence_level,
            'improvement_from_control': improvement_from_control,
        }
        variants_data.append((variant_name, variant_data))
    data.update({'variants': variants_data})
    return data


def _experiments_data():
    experiments_data = {}
    for experiment in backend.all_experiments():
        experiments_data.update({experiment.name:
                                 _experiment_data(experiment)})
    return experiments_data


@login_required
@permission_required('xray.view_experiments', raise_exception=True)
def experiments(request):
    cleaver = request.environ['cleaver']
    tracker.web_event('view_experiments', cleaver.identity)
    tracker.web_event('active', cleaver.identity)
    return render_to_response(
        "xray/experiments.html",
        {"experiments_data": _experiments_data()},
        context_instance=RequestContext(request))


@login_required
@permission_required('xray.view_experiments', raise_exception=True)
def experiments_json(request):
    cleaver = request.environ['cleaver']
    tracker.web_event('view_experiments_json', cleaver.identity)
    tracker.web_event('active', cleaver.identity)
    return HttpResponse(json.dumps(_experiments_data(), cls=DjangoJSONEncoder),
                        mimetype="application/json")


def _events_dates(events):
    dates = {'days': False, 'weeks': False, 'months': False}
    for event in events:
        if 'W' in event.date:
            dates['weeks'] = True
        elif len(event.date.split('-')) == 3:
            dates['days'] = True
        elif len(event.date.split('-')) == 2:
            dates['months'] = True
        else:
            raise RuntimeError('Unknown event date')
    return [k for k, v in dates.items() if v]


@login_required
@permission_required('xray.view_events', raise_exception=True)
def events(request):
    events = tracker.get_events()

    web_events = [e for e in events if e.kind == 'web']
    sms_events = [e for e in events if e.kind == 'sms']

    web_dates = _events_dates(web_events)
    sms_dates = _events_dates(sms_events)

    cleaver = request.environ['cleaver']
    tracker.web_event('view_events', cleaver.identity)
    tracker.web_event('active', cleaver.identity)

    return render_to_response(
        "xray/events_form.html",
        {"web_events": set([(e.name, e.display) for e in web_events]),
         "web_dates": web_dates,
         "sms_events": set([(e.name, e.display) for e in sms_events]),
         "sms_dates": sms_dates},
        context_instance=RequestContext(request))


@login_required
@permission_required('xray.view_events', raise_exception=True)
def events_data(request):
    event_kind = request.GET.get('event_kind')
    assert event_kind in ['web', 'sms']

    select1 = request.GET.get('%s_select1' % event_kind)
    select1 = "xray:%s:%s" % (event_kind, select1)

    select2 = request.GET.get('%s_select2' % event_kind)
    select2 = "xray:%s:%s" % (event_kind, select2)

    time_group = request.GET.get('%s_time_group' % event_kind)

    data = cohort.get_dates_data(select1=select1,
                                 select2=select2,
                                 time_group=time_group,
                                 system='rapidsms-xray')

    """
    column_counter = defaultdict(int)
    column_averages = []
    for column in xrange(2, 15):
        for row in data:
            if row[column]:
                column_counter['%d-count' % column] += 1
                column_counter['%d-total' % column] += row[column]
        if column_counter['%d-count' % column] > 0:
            column_averages.append(float(column_counter['%d-total' % column]) /
                                   column_counter['%d-count' % column])
        else:
            column_averages.append('')
    """

    cleaver = request.environ['cleaver']
    tracker.web_event('view_events_data', cleaver.identity)
    tracker.web_event('active', cleaver.identity)

    return render_to_response(
        "xray/events_data.html",
        {"dates_data": data,
         "time_group": time_group,
         "select1": select1,
         "select2": select2,
         #"averages": column_averages,
         "as_percent": True},
        context_instance=RequestContext(request))



from string import ascii_lowercase

from django import template

register = template.Library()

letters = [l.upper() for l in ascii_lowercase]


@register.filter
def alphabet(index):
    return letters[index]


@register.filter
def cell_style(prct):
    if prct:
        prct = float(prct)
        color = (" color: hsla(200, 100%%, 0%%, %.2f);" %
                 ((prct / 100.0) + 0.5))
        background = ("background-color: hsla(200, 80%%, 50%%, %.2f);" %
                      (prct / 100.0))
        return background + color
    return ""


# the following is from: http://djangosnippets.org/snippets/2147/

class RangeNode(template.Node):
    def __init__(self, parser, range_args, context_name):
        self.template_parser = parser
        self.range_args = range_args
        self.context_name = context_name

    def render(self, context):

        resolved_ranges = []
        for arg in self.range_args:
            compiled_arg = self.template_parser.compile_filter(arg)
            resolved_ranges.append(compiled_arg.resolve(context,
                                                        ignore_failures=True))
        context[self.context_name] = range(*resolved_ranges)
        return ""


@register.tag
def mkrange(parser, token):
    """
    Accepts the same arguments as the 'range' builtin and creates
    a list containing the result of 'range'.

    Syntax:
        {% mkrange [start,] stop[, step] as context_name %}

    For example:
        {% mkrange 5 10 2 as some_range %}
        {% for i in some_range %}
          {{ i }}: Something I want to repeat\n
        {% endfor %}

    Produces:
        5: Something I want to repeat
        7: Something I want to repeat
        9: Something I want to repeat
    """

    tokens = token.split_contents()
    fnctl = tokens.pop(0)

    def error():
        raise template.TemplateSyntaxError,\
            ("%s accepts the syntax: "
             "{%% %s [start,] stop[, step] as context_name %%}, "
             "where 'start', 'stop' and 'step' must all "
             "be integers." % (fnctl, fnctl))

    range_args = []
    while True:
        if len(tokens) < 2:
            error()

        token = tokens.pop(0)

        if token == "as":
            break

        range_args.append(token)

    if len(tokens) != 1:
        error()

    context_name = tokens.pop()

    return RangeNode(parser, range_args, context_name)



from cleaver.identity import CleaverIdentityProvider

import rolodex


class WebIdentityProvider(CleaverIdentityProvider):
    def __init__(self):
        self.rolodex = rolodex.Rolodex()

    def get_identity(self, environ):
        identities = self.rolodex.lookup_browser(environ)
        environ['UID'] = identities.get('uid')
        environ['BID'] = identities.get('bid')
        environ['SID'] = identities.get('sid')
        if identities.get('uid') is not None:
            environ.update({'cleaver_id': identities['uid']})
            return identities['uid']
        if identities.get('bid') is not None:
            environ.update({'cleaver_id': identities['bid']})
            return identities['bid']
        environ.update({'cleaver_id': identities['sid']})
        return identities['sid']


class SMSIdentityProvider(CleaverIdentityProvider):
    def __init__(self):
        self.rolodex = rolodex.Rolodex()

    def get_identity(self, message_fields):
        # create hash of backend name and identity
        identities = self.rolodex.lookup_msisdn(message_fields['connection'].identity)
        # TODO use uid if available?
        cleaver_id = identities['mid']
        message_fields.update({'cleaver_id': cleaver_id,
                               'uid': identities.get('uid'),
                               'mid': identities.get('mid')})
        return cleaver_id




def web_experiments(request):
    cleaver = request.environ['cleaver']

    experiments = {}

    experiments['web_xray_dashboard_experiments_icon_variant'] = cleaver(
        'web_xray_dashboard_experiments_icon',
        ('beaker', 'icon-beaker'),
        ('stethoscope', 'icon-stethoscope')
    )

    experiments['web_xray_dashboard_phone_icon_variant'] = cleaver(
        'web_xray_dashboard_phone_icon',
        ('mobile-phone', 'icon-mobile-phone'),
        ('phone', 'icon-phone'),
        ('phone-sign', 'icon-phone-sign')
    )

    experiments['web_xray_dashboard_web_icon_variant'] = cleaver(
        'web_xray_dashboard_web_icon',
        ('laptop', 'icon-laptop'),
        ('desktop', 'icon-desktop')
    )

    if request.path == '/xray/events/':
        cleaver.score('web_xray_dashboard_phone_icon')
        cleaver.score('web_xray_dashboard_web_icon')

    if request.path == '/xray/experiments/':
        cleaver.score('web_xray_dashboard_experiments_icon')

    return experiments



def memoize(f):
    """ Memoization decorator for a function taking one or more arguments. """
    class memodict(dict):
        def __getitem__(self, *key):
            return dict.__getitem__(self, key)

        def __missing__(self, key):
            ret = self[key] = f(*key)
            return ret

    return memodict().__getitem__



import hashlib

import redis
import phonenumbers

from django.conf import settings

SessionStore = __import__(settings.SESSION_ENGINE,
                          fromlist=['']).SessionStore

# thanks to:
# http://mobiforge.com/forum/running/analytics/
# stats-and-unique-vistors-different-mobile
# and
# http://mobiforge.com/developing/blog/useful-x-headers
headers_for_fingerprint = ['USER-AGENT', 'HOST', 'ACCEPT', 'ACCEPT-LANGUAGE',
                           'ACCEPT-CHARSET', 'X-REAL-IP', 'X-FORWARDED-HOST',
                           'HTTP-X-REAL-IP', 'HTTP_X_FORWARDED_FOR',
                           'HTTP_USER_AGENT', 'HTTP_DNT', 'HTTP_ACCEPT',
                           'HTTP_ACCEPT_LANGUAGE', 'HTTP_ACCEPT_ENCODING',
                           'X-FORWARDED-SERVER', 'X-FORWARDED-FOR',
                           'X-UP-SUBNO', 'X-NOKIA-MSISDN',
                           'X-UP-CALLING-LINE-ID', 'X-HTS-CLID',
                           'X-H3G-MSISDN', 'X-NX-CLID', 'X-ACCESS-SUBNYM',
                           'X-ORANGE-ID', 'MSISDN', 'X-WAP-PROFILE',
                           'X-WAP-PROFILE-DIFF', 'X-APN-ID',
                           'X-DRUTT-DEVICE-ID', 'X-DRUTT-PORTAL-USER-ID',
                           'X-DRUTT-PORTAL-USER-MSISDN', 'X-GGSNIP',
                           'X-JPHONE-COLOR', 'X-JPHONE-DISPLAY',
                           'X-NETWORK-INFO', 'X-OS-PREFS', 'X-NOKIA-ALIAS',
                           'X-NOKIA-BEARER', 'X-NOKIA-IMSI', 'X-NOKIA-MSISDN',
                           'X-OPERAMINI-PHONE', 'X-ORIGINAL-USER-AGENT',
                           'X-IMSI', 'X-MSISDN']


class Hashabledict(dict):
    # from http://stackoverflow.com/a/16162138
    def __hash__(self):
        return hash((frozenset(self), frozenset(self.itervalues())))


class Rolodex(object):
    # 'mobile id'
    # mid = hashlib.md5(e164-formatted-msisdn).hexdigest()

    # find uid from phone number
    # (string) 'uid:{{ e164 }}' => uid
    # find uid from mid
    # (string) 'uid:{{ mid }}' => uid
    # find uid from bid
    # (string) 'uid:{{ bid }}' => uid

    # find phone number from mid
    # (string) 'e164:{{ mid }}' => e164

    # find count of phone activity
    # (sortedset) midCounts => (mid,n) where n is number of lookups

    # find count of browser activity
    # (sortedset) bidCounts => (bid,n) where n is number of lookups

    # find browsers used by mid
    # (set) 'bid:{{ mid }}' => (bid,...)

    # find uid and mid from browser
    # (hash) 'bid:{{ bid }}' => {'uid': uid, 'mid': mid}

    # seen mids
    # (set) midSeen => (mid,...)
    # registered mids
    # (set) midRegistered => (mid,...)

    # find mids used by uid
    # (set) 'mids:{{ uid }}' => (mid,...)

    # find browsers used by uid
    # (set) 'bids:{{ uid }}' => (bid,...)

    # invalid msisdn for country, scored by frequency
    # (sorted set) 'invalid:msisdn:{{ country }}' => ((1234, 22),(5678, 8),...)

    def __init__(self, host='localhost', port=6379, db=4, prefix="rolodex",
                 country='UG'):
        assert country in phonenumbers.SUPPORTED_REGIONS
        self.prefix = prefix
        # TODO allow list of countries?
        self.country = country
        self.redis = redis.Redis(host=host, port=port, db=db)

    def xid(self, hashable):

        # unique ids must be smaller than 2**32 -1 in order
        # to use redis' bitmaps to store event data.
        # and the smaller, the better, as it will use less RAM.
        # so, for each unique md5 hash, issue an integer

        hashed = hashlib.md5(hashable).hexdigest()

        existing = self.redis.get('%s:xid:%s' % (self.prefix, hashed))
        if existing:
            return existing

        # increment counter
        integer_id = self.redis.incrby('%s:xray_identity_counter' %
                                       self.prefix, 1)
        # set maps
        self.redis.set('%s:xid:%s' % (self.prefix, hashed), integer_id)
        self.redis.set('%s:xid:%s' % (self.prefix, integer_id), hashed)
        return integer_id

    def format_msisdn(self, msisdn=None):
        """ given a msisdn, return in E164 format """
        assert msisdn is not None
        num = phonenumbers.parse(msisdn, self.country)
        is_valid = phonenumbers.is_valid_number(num)
        if not is_valid:
            #raise RuntimeError("%s is not a valid number for %s"
            #                   % (msisdn, self.country))
            # create or increment count of  invalid number
            self.redis.zadd('%s:invalid:msisdn:%s' %
                            (self.prefix, self.country), msisdn, 1.0)
            return msisdn
        return phonenumbers.format_number(num,
                                          phonenumbers.PhoneNumberFormat.E164)

    def mid_for_msisdn(self, msisdn=None):
        assert msisdn is not None
        return self.xid(self.format_msisdn(msisdn))

    def _seen_mid_registered(self, mid):
        self.redis.sadd('%s:midRegistered' % self.prefix, mid)

    def _seen_mid(self, mid, e164):
        self.redis.zincrby('%s:midCounts' % self.prefix, mid, 1.0)
        self.redis.set('%s:e164:%s' % (self.prefix, mid), e164)
        self.redis.sadd('%s:midSeen' % self.prefix, mid)

    def uid_for_mid(self, mid):
        return self.redis.get('%s:uid:%s' % (self.prefix, mid))

    def mids_for_uid(self, uid):
        return self.redis.smembers('%s:mids:%s' % (self.prefix, uid))

    def _seen_bid(self, bid, uid, sid):
        self.redis.zincrby('%s:bidCounts' % self.prefix, bid)
        if uid is not None:
            self.redis.hmset('%s:ids:%s' % (self.prefix, bid),
                             {'uid': uid, 'sid': sid})
            # cache only for 2 minutes. if browser activity continues, value
            # will be recached
            self.redis.expire('%s:ids:%s' % (self.prefix, bid), 120)
            return True
        return None

    def ids_for_bid(self, bid):
        if self.redis.hexists('%s:ids:%s' % (self.prefix, bid), 'uid'):
            return self.redis.hgetall('%s:ids:%s' % (self.prefix, bid))
        return None

    def expire_bid(self, bid):
        if self.redis.exists('%s:ids:%s' % (self.prefix, bid)):
            return self.redis.delete('%s:ids:%s' % (self.prefix, bid))
        return None

    def lookup_msisdn(self, msisdn=None):
        assert msisdn is not None
        e164 = self.format_msisdn(msisdn)
        mid = None
        uid = None
        if e164:
            mid = self.xid(e164)
        else:
            # if number cannot be e164 formatted,
            # return hash of string
            # could be an operator's message, etc
            mid = self.xid(msisdn)

        self._seen_mid(mid, e164)

        uid = self.uid_for_mid(mid)
        if uid:
            self._seen_mid_registered(mid)

        return {'mid': mid, 'uid': uid}

    def fingerprint_environ(self, hashable_environ):
        headers_up = dict(((k.upper(), v) for k, v
                           in hashable_environ.iteritems()
                           if isinstance(v, str)))
        header_info = []
        for header in headers_for_fingerprint:
            header_info.append(headers_up.get(header, ''))
        xid = self.xid(''.join(header_info).replace(' ', ''))
        return int(xid)

    def lookup_browser(self, environ):
        bid = None
        uid = None
        sid = None
        hashable_environ = Hashabledict(environ)
        bid = self.fingerprint_environ(hashable_environ)

        if self.ids_for_bid(bid):
            self._seen_bid(bid, uid, sid)
            # add 'bid': bid to cached values and return
            return dict([('bid', bid)] + self.ids_for_bid(bid).items())

        if 'HTTP_COOKIE' in environ:
            cookie = {s.split('=')[0].strip(): s.split('=')[1].strip()
                      for s in environ['HTTP_COOKIE'].split(';')}
            if 'sessionid' in cookie:
                sid = self.xid(cookie['sessionid'])
                environ['SID'] = sid
                session = SessionStore(session_key=cookie['sessionid'])
                if session.exists(cookie['sessionid']):
                    session.load()
                    uid = self.xid(str(session.get('_auth_user_id')))
        self._seen_bid(bid, uid, sid)
        return {'uid': uid, 'bid': bid, 'sid': sid}



from django.dispatch import receiver
from django.contrib.auth.signals import user_logged_in
from django.contrib.auth.signals import user_logged_out

from .events import Tracker
from . import rolodex

tracker = Tracker()


@receiver(user_logged_in)
def receive_logged_in(**kwargs):
    request = kwargs.get("request")
    cleaver = request.environ['cleaver']

    tracker.web_event('logged_in', cleaver.identity)
    tracker.web_event('active', cleaver.identity)


@receiver(user_logged_out)
def receive_logged_out(**kwargs):
    request = kwargs.get("request")
    cleaver = request.environ['cleaver']

    tracker.web_event('logged_out', cleaver.identity)
    rolo = rolodex.Rolodex()
    rolo.expire_bid(request.environ['BID'])



from django.conf.urls.defaults import patterns
from django.conf.urls.defaults import url

from . import views

urlpatterns = patterns('',
                       url(r'^$', views.index, name='xray_dashboard'),
                       url(r'^experiments/$', views.experiments,
                           name='xray_experiments'),
                       url(r'^api/experiments/$',
                           views.experiments_json,
                           name='xray_experiments_json'),
                       url(r'^events/data/$', views.events_data,
                           name='xray_events_data'),
                       url(r'^events/$', views.events, name='xray_events'),
                       )



from collections import namedtuple
import bitmapist

Event = namedtuple('Event', 'prefix, kind, name, date, display')


class Tracker(object):
    def __init__(self, name='rapidsms-xray', prefix='xray', host='localhost',
                 port=6379, db=3):
        self.prefix = prefix
        bitmapist.setup_redis(name, host, port, db=db)

    def _event(self, event_name, identifier, system='rapidsms-xray',
               now=None, track_hourly=None):
        bitmapist.mark_event(event_name=event_name, uuid=identifier,
                             system=system, now=now, track_hourly=track_hourly)

    def web_event(self, event_name, identifier, system='rapidsms-xray',
                  now=None, track_hourly=None):
        event_name = '%s:web:%s' % (self.prefix, event_name)
        #print ('EVENT %s %s' % (event_name, identifier))
        self._event(event_name, identifier, system, now, track_hourly)

    def sms_event(self, event_name, identifier, system='rapidsms-xray',
                  now=None, track_hourly=None):
        event_name = '%s:sms:%s' % (self.prefix, event_name)
        #print ('EVENT %s %s' % (event_name, identifier))
        self._event(event_name, identifier, system, now, track_hourly)

    def get_events(self):
        redis = bitmapist.get_redis('rapidsms-xray')
        all_keys = redis.keys('trackist_%s:*' % self.prefix)
        events = []
        for key in all_keys:
            prefix, kind, event_info = key.split(':')
            event_info_list = event_info.split('_')
            event_date = event_info_list.pop()
            event_display = ' '.join(event_info_list)
            event_name = '_'.join(event_info_list)
            events.append(Event._make((prefix, kind, event_name, event_date,
                                       event_display)))
        return events


# TODO pypi has old version without  bitmapist.MixinEventsMisc (which replaces
# MixinEventsMarked)
class WebWeekEvents(bitmapist.MixinCounts, bitmapist.MixinContains,
                    bitmapist.MixinEventsMarked):
    def __init__(self, event_name, year, week, system='rapidsms-xray',
                 prefix='xray'):
        self.system = system
        self.prefix = prefix
        event_name = '%s:web:%s' % (self.prefix, event_name)
        self.redis_key = bitmapist._prefix_key(event_name,
                                               'W%s-%s' % (year, week))


class SMSWeekEvents(bitmapist.MixinCounts, bitmapist.MixinContains,
                    bitmapist.MixinEventsMarked):
    def __init__(self, event_name, year, week, system='rapidsms-xray',
                 prefix='xray'):
        self.system = system
        self.prefix = prefix
        event_name = '%s:sms:%s' % (self.prefix, event_name)
        self.redis_key = bitmapist._prefix_key(event_name,
                                               'W%s-%s' % (year, week))


class WebMonthEvents(bitmapist.MixinCounts, bitmapist.MixinContains,
                     bitmapist.MixinEventsMarked):
    def __init__(self, event_name, year, month, system='rapidsms-xray',
                 prefix='xray'):
        self.system = system
        self.prefix = prefix
        event_name = '%s:web:%s' % (self.prefix, event_name)
        self.redis_key = bitmapist._prefix_key(event_name,
                                               '%s-%s' % (year, month))


class SMSMonthEvents(bitmapist.MixinCounts, bitmapist.MixinContains,
                     bitmapist.MixinEventsMarked):
    def __init__(self, event_name, year, month, system='rapidsms-xray',
                 prefix='xray'):
        self.system = system
        self.prefix = prefix
        event_name = '%s:sms:%s' % (self.prefix, event_name)
        self.redis_key = bitmapist._prefix_key(event_name,
                                               '%s-%s' % (year, month))


class WebDayEvents(bitmapist.MixinCounts, bitmapist.MixinContains,
                   bitmapist.MixinEventsMarked):
    def __init__(self, event_name, year, month, day, system='rapidsms-xray',
                 prefix='xray'):
        self.system = system
        self.prefix = prefix
        event_name = '%s:web:%s' % (self.prefix, event_name)
        self.redis_key = bitmapist._prefix_key(event_name,
                                               '%s-%s-%s' % (year, month, day))


class SMSDayEvents(bitmapist.MixinCounts, bitmapist.MixinContains,
                   bitmapist.MixinEventsMarked):
    def __init__(self, event_name, year, month, day, system='rapidsms-xray',
                 prefix='xray'):
        self.system = system
        self.prefix = prefix
        event_name = '%s:sms:%s' % (self.prefix, event_name)
        self.redis_key = bitmapist._prefix_key(event_name,
                                               '%s-%s-%s' % (year, month, day))


class WebHourEvents(bitmapist.MixinCounts, bitmapist.MixinContains,
                    bitmapist.MixinEventsMarked):
    def __init__(self, event_name, year, month, day, hour,
                 system='rapidsms-xray', prefix='xray'):
        self.system = system
        self.prefix = prefix
        event_name = '%s:web:%s' % (self.prefix, event_name)
        self.redis_key = bitmapist._prefix_key(event_name,
                                               '%s-%s-%s-%s' %
                                               (year, month, day, hour))


class SMSHourEvents(bitmapist.MixinCounts, bitmapist.MixinContains,
                    bitmapist.MixinEventsMarked):
    def __init__(self, event_name, year, month, day, hour,
                 system='rapidsms-xray', prefix='xray'):
        self.system = system
        self.prefix = prefix
        event_name = '%s:sms:%s' % (self.prefix, event_name)
        self.redis_key = bitmapist._prefix_key(event_name,
                                               '%s-%s-%s-%s' %
                                               (year, month, day, hour))



from __future__ import unicode_literals
from wsgiref.util import setup_testing_defaults

from .base import XrayTestBase
from .headers import headers
import sys


def debug(msg):
    sys.stdout.write(' %s ' % str(msg))
    sys.stdout.flush()


class RolodexTest(XrayTestBase):

    def setUp(self):
        super(RolodexTest, self).setUp()


class RolodexWebIDTest(RolodexTest):

    def test_lookup_browser(self):
        # lookup_browser returns same fingerprint
        environ_one = headers[0]
        setup_testing_defaults(environ_one)

        browser_one_ids = self.rolodex.lookup_browser(environ_one)
        assert browser_one_ids['bid'] == 1

        browser_count = self.rolodex_db.zrange('%s:bidCounts' % self.prefix,
                                               0, 1000)
        assert len(browser_count) == 1
        browser_one_seen = self.rolodex_db.zscore('%s:bidCounts' % self.prefix,
                                                  browser_one_ids['bid'])
        assert browser_one_seen == 1

        # lookup_browser issues only one xid for same headers
        browser_one_ids2 = self.rolodex.lookup_browser(environ_one)
        assert browser_one_ids['bid'] == browser_one_ids2['bid']

        browser_count = self.rolodex_db.zrange('%s:bidCounts' % self.prefix,
                                               0, 1000)
        assert len(browser_count) == 1

        browser_one_seen = self.rolodex_db.zscore('%s:bidCounts' % self.prefix,
                                                  browser_one_ids['bid'])
        assert browser_one_seen == 2

        # bidcounts is incremented
        environ_two = headers[1]
        setup_testing_defaults(environ_two)
        browser_two_ids = self.rolodex.lookup_browser(environ_two)

        assert browser_two_ids['bid'] != browser_one_ids['bid']
        assert browser_two_ids['bid'] == 2
        browser_count = self.rolodex_db.zrange('%s:bidCounts' % self.prefix,
                                               0, 1000)
        assert len(browser_count) == 2

        browser_one_seen = self.rolodex_db.zscore('%s:bidCounts' % self.prefix,
                                                  browser_one_ids['bid'])
        assert browser_one_seen == 2
        browser_two_seen = self.rolodex_db.zscore('%s:bidCounts' % self.prefix,
                                                  browser_two_ids['bid'])
        assert browser_two_seen == 1
        # TODO :ids: is set
        # TODO :ids: expires

        # TODO same browser logs in, issued different xid
        # TODO log in with another browers, gets same xid
        pass


class RolodexSMSIDTest(RolodexTest):
    # TODO lookup_msisdn returns same fingerprint
    # TODO lookup_msisdn issues only one xid
    # TODO mappings are created
    pass



from cleaver.base import Cleaver
from cleaver.identity import CleaverIdentityProvider
from cleaver.backend import CleaverBackend
from cleaver.backend.redis import RedisBackend

from rapidsms.apps.base import AppBase

from .identity import SMSIdentityProvider
from .events import Tracker


class App(AppBase):

    def __init_cleaver(self, cleaver_identity, cleaver_backend):
        if not isinstance(cleaver_identity, CleaverIdentityProvider) and \
                not callable(cleaver_identity):
            raise RuntimeError(
                '%s must be callable or implement '
                'cleaver.identity.CleaverIdentityProvider' % cleaver_identity
            )
        if not isinstance(cleaver_backend, CleaverBackend):
            raise RuntimeError(
                '%s must implement cleaver.backend.CleaverBackend' %
                cleaver_backend
            )
        self._cleaver_backend = cleaver_backend
        self._cleaver_identity = cleaver_identity

    def __init__(self, router):
        # TODO use project config for RedisBackend and Tracker?
        self.__init_cleaver(SMSIdentityProvider(),
                            RedisBackend())
        # initialize event tracker
        self.tracker = Tracker()

    def filter(self, message):

        # add connection to message metadata because
        # only the metadata (not the message object)
        # is passed to cleaver
        message.fields.update({'connection': message.connection})

        cleaver = Cleaver(
            message.fields,
            self._cleaver_identity,
            self._cleaver_backend
        )
        message.fields.update({'cleaver': cleaver})

        # Mark the visitor as a human
        self._cleaver_backend.mark_human(cleaver.identity)

        # If the visitor has been assigned any experiment variants,
        # tally their participation.
        for e in self._cleaver_backend.all_experiments():
            variant = self._cleaver_backend.get_variant(
                cleaver.identity,
                e.name
            )
            if variant:
                self._cleaver_backend.mark_participant(cleaver.identity,
                                                       e.name, variant)

        self.tracker.sms_event('active', cleaver.identity)

    def handle(self, message):

        if message.text.startswith('wat'):
            cleaver = message.fields['cleaver']
            # test a few responses to 'wat'
            response = cleaver(
                'sms_wat_response',
                ('foo', 'foo'),
                ('bar', 'bar'),
                ('baz', 'baz')
            )
            message.respond(response)
            return True

        if message.text.startswith('huh'):
            # score conversion if user sends 'huh'
            # note this conversion could be scored in
            # another app or handler because the cleaver
            # object is added to the message's fields
            # property during the `filter` phase
            cleaver = message.fields['cleaver']
            cleaver.score('sms_wat_response')



"""RapidSMS app for web and SMS split test experiments & event tracking."""

# import signal receivers here so they are sure
# to be registered before any signals are emitted
# ignore ImportError that may be thrown when running setup.py
try:
    import listeners
except ImportError:
    pass

__version_info__ = {
    'major': 0,
    'minor': 5,
    'micro': 2,
    'releaselevel': 'beta',
    'serial': 0
}


def get_version(short=False):
    assert __version_info__['releaselevel'] in ('alpha', 'beta', 'final')
    vers = ["%(major)i.%(minor)i" % __version_info__, ]
    if __version_info__['micro']:
        vers.append(".%(micro)i" % __version_info__)
    if __version_info__['releaselevel'] != 'final' and not short:
        vers.append('%s%i' % (__version_info__['releaselevel'][0],
                              __version_info__['serial']))
    return ''.join(vers)

__version__ = get_version()



from setuptools import setup, find_packages


requirements = [line.strip('\n') for line in open('requirements/base.txt').readlines()]

setup(
    name='rapidsms-xray',
    version=__import__('xray').__version__,
    author='Evan Wheeler',
    author_email='evan@leapfrog.io',
    packages=find_packages(),
    include_package_data=True,
    dependency_links = ['http://github.com/ewheeler/cleaver/tarball/xray#egg=cleaver-0.1.5dev'],
    install_requires=requirements,
    url='https://github.com/ewheeler/rapidsms-xray/',
    license='BSD',
    description=u' '.join(__import__('xray').__doc__.splitlines()).strip(),
    classifiers=[
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
    ],
    long_description=open('README.rst').read(),
    test_suite='runtests.runtests',
    zip_safe=False,
)



headers = [\
{
   "HTTP_X_WAP_PROFILE": "\"http://nds1.nds.nokia.com/uaprof/NX2-00r100.xml\", \"1-4xx28zdY1/BbVIZPpYDH/A==\", \"2-XnrTOLDzBJdZHN2vSasoNA==\", \"3-HC5l5j+eQ9tPRpdhsseJIQ==\"",
   "HTTP_COOKIE": "Bearer-Type=w-TCP;wtls-security-level=none;network-access-type=CSD",
   "SERVER_SOFTWARE": "Apache/2.2.17 (Ubuntu)",
   "SCRIPT_NAME": "",
   "QUERY_STRING": "code=dddfb329-9124-4e56-b02f-b430b0716de8&email=talk2femi09%40yahoo.com",
   "HTTP_X_WAP_PROFILE_DIFF": "1; <?xml version=\"1.0\"?><rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\" xmlns:prf=\"http://www.wapforum.org/UAPROF/ccppschema-19991014#\"><!-- browser vendor site: Default description of properties --><rdf:Description><prf:CcppAccept><rdf:Bag><rdf:li>application/vnd.wap.wmlscriptc</rdf:li><rdf:li>text/vnd.wap.wml</rdf:li><rdf:li>application/vnd.wap.xhtml+xml</rdf:li><rdf:li>application/xhtml+xml</rdf:li><rdf:li>text/xml</rdf:li><rdf:li>text/html</rdf:li><rdf:li>text/css</rdf:li><rdf:li>multipart/mixed</rdf:li><rdf:li>*/*</rdf:li><rdf:li>text/vnd.wap.wmlscript</rdf:li></rdf:Bag></prf:CcppAccept></rdf:Description></rdf:RDF>",
   "PATH_INFO": "/",
   "SERVER_PROTOCOL": "HTTP/1.1",
   "HTTP_X_NOKIA_MUSICSHOP_BEARER": "GPRS/3G",
   "HTTP_X_FORWARDED_PORT": "80",
   "HTTP_ACCEPT_CHARSET": "*",
   "HTTP_USER_AGENT": "NokiaX2-00/5.0 (04.90) Profile/MIDP-2.1 Configuration/CLDC-1.1 Mozilla/5.0 AppleWebKit/420+ (KHTML, like Gecko) Safari/420+",
   "HTTP_CONNECTION": "keep-alive",
   "SERVER_NAME": "example.com",
   "REMOTE_ADDR": "10.210.247.150",
   "HTTP_X_NOKIA_MUSICSHOP_VERSION": "1.2.1",
   "SERVER_PORT": "80",
   "SERVER_ADDR": "10.72.87.133",
   "remote_ip": "41.206.15.18",
   "format": "application/vnd.wap.xhtml+xml",
   "HTTP_X_FORWARDED_PROTO": "http",
   "SERVER_ADMIN": "[no address given]",
   "SCRIPT_URI": "http://example.com/",
   "HTTP_DRM_VERSION": "2.0",
   "HTTP_HOST": "example.com",
   "SCRIPT_URL": "/",
   "HTTP_COOKIE2": "$Version=\"1\"",
   "REQUEST_URI": "/?code=dddfb329-9124-4e56-b02f-b430b0716de8&email=talk2femi09%40yahoo.com",
   "HTTP_ACCEPT": "application/vnd.wap.wmlscriptc, text/vnd.wap.wml, application/vnd.wap.xhtml+xml, application/xhtml+xml, text/xml, text/html, text/css, multipart/mixed, */*, text/vnd.wap.wmlscript",
   "HTTP_WAP_CONNECTION": "Stack-Type=HTTP",
   "url": "http://example.com/?code=dddfb329-9124-4e56-b02f-b430b0716de8&email=talk2femi09%40yahoo.com",
   "REMOTE_PORT": "41180",
   "HTTP_ACCEPT_LANGUAGE": "en",
   "HTTP_X_FORWARDED_FOR": "41.206.15.18",
   "HTTP_X_REQUEST_START": "t=1313615637103298",
   "REQUEST_METHOD": "GET",
   "DOCUMENT_ROOT": "/var/www"
},
{
   "HTTP_REFERER": "http://example.com/",
   "SERVER_SOFTWARE": "Apache/2.2.17 (Ubuntu)",
   "SCRIPT_NAME": "",
   "HTTP_X_FORWARDED_PORT": "80",
   "PATH_INFO": "/",
   "SERVER_PROTOCOL": "HTTP/1.1",
   "QUERY_STRING": "",
   "HTTP_ACCEPT_CHARSET": "iso-8859-1, utf-8, utf-16, *;q=0.1",
   "HTTP_USER_AGENT": "Opera/9.80 (J2ME/MIDP; Opera Mini/4.2.14881/25.729; U; en) Presto/2.5.25 Version/10.54",
   "HTTP_CONNECTION": "keep-alive",
   "HTTP_COOKIE": "_Members_session=BAh7BiIPc2Vzc2lvbl9pZCIlZTA1NDY4YjFjYTZiMDczMTAzM2U4MWYwYzhmNzZjNDI%3D--f5a7a61676e61cf449d83e3d6c3e0e12b5364917",
   "SERVER_NAME": "example.com",
   "HTTP_X_OPERAMINI_PHONE_UA": "NokiaX2-00",
   "SERVER_PORT": "80",
   "SERVER_ADDR": "10.72.11.51",
   "remote_ip": "41.206.15.18",
   "REMOTE_PORT": "15897",
   "format": "text/html",
   "HTTP_X_FORWARDED_PROTO": "http",
   "SERVER_ADMIN": "[no address given]",
   "SCRIPT_URI": "http://example.com/",
   "HTTP_HOST": "example.com",
   "SCRIPT_URL": "/",
   "HTTP_X_OPERAMINI_FEATURES": "advanced, file_system, folding",
   "HTTP_COOKIE2": "$Version=1",
   "REQUEST_URI": "/",
   "HTTP_ACCEPT": "text/html, application/xml;q=0.9, application/xhtml+xml, image/png, image/jpeg, image/gif, image/x-xbitmap, */*;q=0.1",
   "HTTP_X_OPERAMINI_PHONE": "Nokia # X2-00",
   "url": "http://example.com/",
   "REMOTE_ADDR": "10.210.247.150",
   "HTTP_ACCEPT_LANGUAGE": "en",
   "HTTP_X_FORWARDED_FOR": "41.206.15.18, 82.145.210.98",
   "HTTP_X_REQUEST_START": "t=1313615754667339",
   "REQUEST_METHOD": "GET",
   "DOCUMENT_ROOT": "/var/www/",
   "HTTP_ACCEPT_ENCODING": "deflate, gzip, x-gzip, identity, *;q=0"
},
{
   "HTTP_REFERER": "http://gowaping.com/site_90.xhtml?get-id=738",
   "SERVER_SOFTWARE": "Apache/2.2.17 (Ubuntu)",
   "SCRIPT_NAME": "",
   "HTTP_X_FORWARDED_PORT": "80",
   "PATH_INFO": "/",
   "SERVER_PROTOCOL": "HTTP/1.1",
   "QUERY_STRING": "cid=BC_NG_ATS3_EN_0",
   "HTTP_ACCEPT_CHARSET": "iso-8859-1, utf-8; q=0.7, *; q=0.7",
   "HTTP_USER_AGENT": "Mozilla/5.0 (SymbianOS/9.1; U; [en]; SymbianOS/9.1 Series60/3.0) AppleWebKit/413 (KHTML, like Gecko) Safari/413",
   "HTTP_CONNECTION": "keep-alive",
   "SERVER_NAME": "example.com",
   "REMOTE_ADDR": "10.210.247.150",
   "SERVER_PORT": "80",
   "REQUEST_URI": "/?cid=BC_NG_ATS3_EN_0",
   "SERVER_ADDR": "10.240.110.124",
   "remote_ip": "41.190.3.221",
   "format": "text/html",
   "HTTP_X_FORWARDED_PROTO": "http",
   "SERVER_ADMIN": "[no address given]",
   "SCRIPT_URI": "http://example.com/",
   "HTTP_HOST": "example.com",
   "SCRIPT_URL": "/",
   "HTTP_ACCEPT": "text/html,application/xhtml+xml,text/css, */*",
   "url": "http://example.com/?cid=BC_NG_ATS3_EN_0",
   "REMOTE_PORT": "22566",
   "HTTP_ACCEPT_LANGUAGE": "en;q=1.0,ar;q=0.5,fr;q=0.5,de;q=0.5,es;q=0.5,it;q=0.5",
   "HTTP_X_FORWARDED_FOR": "41.190.3.221",
   "HTTP_X_REQUEST_START": "t=1313616134527736",
   "REQUEST_METHOD": "GET",
   "DOCUMENT_ROOT": "/var/www/",
   "HTTP_ACCEPT_ENCODING": "gzip, deflate, x-gzip, identity; q=0.9"
},
{
   "HTTP_X_WAP_PROFILE": "\"http://www.htcmms.com.tw/gen/Herald-1.0.xml\"",
   "HTTP_REFERER": "http://ad.mosko.mobi/",
   "SERVER_SOFTWARE": "Apache/2.2.17 (Ubuntu)",
   "SCRIPT_NAME": "",
   "HTTP_X_FORWARDED_PORT": "80",
   "PATH_INFO": "/",
   "SERVER_PROTOCOL": "HTTP/1.1",
   "QUERY_STRING": "cid=BC_NG_ATS3_EN_0",
   "HTTP_UA_COLOR": "color16",
   "HTTP_USER_AGENT": "Mozilla/4.0 (compatible; MSIE 6.0; Windows CE; PPC)",
   "HTTP_CONNECTION": "keep-alive",
   "SERVER_NAME": "example.com",
   "REMOTE_ADDR": "10.210.247.150",
   "SERVER_PORT": "80",
   "REQUEST_URI": "/?cid=BC_NG_ATS3_EN_0",
   "SERVER_ADDR": "10.72.11.51",
   "remote_ip": "41.206.11.37",
   "HTTP_UA_OS": "Windows CE (Pocket PC) - Version 5.2",
   "format": "application/vnd.wap.xhtml+xml",
   "HTTP_X_FORWARDED_PROTO": "http",
   "SERVER_ADMIN": "[no address given]",
   "SCRIPT_URI": "http://example.com/",
   "HTTP_HOST": "example.com",
   "SCRIPT_URL": "/",
   "HTTP_UA_CPU": "ARM",
   "HTTP_ACCEPT": "*/*",
   "url": "http://example.com/?cid=BC_NG_ATS3_EN_0",
   "REMOTE_PORT": "16884",
   "HTTP_ACCEPT_LANGUAGE": "en-us",
   "HTTP_UA_VOICE": "TRUE",
   "HTTP_UA_PIXELS": "320x240",
   "HTTP_X_REQUEST_START": "t=1313615865674202",
   "REQUEST_METHOD": "GET",
   "HTTP_X_FORWARDED_FOR": "41.206.11.37",
   "DOCUMENT_ROOT": "/var/www/",
   "HTTP_ACCEPT_ENCODING": "gzip,deflate"
}
]
